@@ -158,6 +158,11 @@ class JupyterConsole(BaseWidget):
             name=self.name,
             workspace_func=self.get_workspace_globals
         )
+        self.extension_manager.fire(
+            'register_subprocess',
+            pid=self.pid,
+            description='jupyter_console:{}'.format(self.name)
+        )
 
     def interrupt(self):
 
@@ -206,7 +211,7 @@ class JupyterConsole(BaseWidget):
             CHANGE_DIR_CMD.get(
                 self._kernel,
                 DEFAULT_CHANGE_DIR_CMD
-            ).format(path=path)
+            ).format(path=path.replace(u'\\', u'\\\\'))
         )
 
     def run_file(self, path):
@@ -215,7 +220,7 @@ class JupyterConsole(BaseWidget):
             RUN_FILE_CMD.get(
                 self._kernel,
                 DEFAULT_RUN_FILE_CMD
-            ).format(path=path)
+            ).format(path=path.replace(u'\\', u'\\\\'))
         )
 
     def run_debug(self, path, breakpoints):
